#![allow(clippy::upper_case_acronyms)]

use clap::{Arg, Command};

// TODO: extract as icepick_coin
mod icepick_coin {
    use clap::Command;
    pub trait Coin {
        /// Return the name for the coin, in lowercase.
        fn coin_name(&self) -> &'static str;

        /// Return a description of the coin.
        fn coin_description(&self) -> &'static str;

        /// Given a Command, add any arguments relevant to the Transfer operation.
        ///
        /// All commands by default have the following values:
        /// * `amount`: The amount of the currency to transfer.
        /// * `to`: The account to transfer the currency to.
        /// * `from_account`: The non-default derivation account to send currency from.
        fn build_transfer_command(&self, _command: Command) -> Result<Command, ()> {
            Err(())
        }

        /// Given a Command, add any arguments relevant to the Stake operation.
        fn build_stake_command(&self, _command: Command) -> Result<Command, ()> {
            Err(())
        }
    }
}

// TODO: extract as icepick_eth
mod eth {
    use super::icepick_coin::Coin;
    use clap::arg;

    pub struct ETH;

    impl Coin for ETH {
        fn coin_name(&self) -> &'static str {
            "eth"
        }

        fn coin_description(&self) -> &'static str {
            "The leading platform for innovative apps and blockchain networks."
        }

        fn build_transfer_command(&self, command: clap::Command) -> Result<clap::Command, ()> {
            Ok(command
               .arg(arg!(--gas <GAS> "Specify custom gas for the transaction."))
               .arg(arg!(--"gas-payer" <ACCOUNT> "Use an alternative derivation account for paying gas.")))
        }

        fn build_stake_command(&self, _command: clap::Command) -> Result<clap::Command, ()> {
            Err(())
        }
    }
}

use icepick_coin::Coin;

fn build_transfer_command() -> Command {
    Command::new("transfer")
        .about("Transfer coins from one wallet to another.")
}

fn build_stake_command() -> Command {
    Command::new("stake")
        .about("Stake coins.")
}

fn build_transfer_subcommand(coin: &dyn Coin) -> Command {
    let name = coin.coin_name();
    let description = coin.coin_description();
    let uppercase_name = name.to_uppercase();
    Command::new(name)
        .about(description)
        .arg(Arg::new("amount").help(format!("Amount of {uppercase_name} to transfer")))
        .arg(Arg::new("to_address").help(format!("Account to send {uppercase_name} to")))
        .arg(
            Arg::new("from_account")
                .help(format!(
                    "Specify a non-default derivation account to send {uppercase_name} from"
                ))
                .required(false),
        )
}

fn build_stake_subcommand(coin: &dyn Coin) -> Command {
    let name = coin.coin_name();
    let description = coin.coin_description();
    let uppercase_name = name.to_uppercase();
    Command::new(name)
        .about(description)
        .arg(Arg::new("amount").help(format!("Amount of {uppercase_name} to stake")))
        .arg(
            Arg::new("from_account")
                .help(format!(
                    "Specify a non-default derivation account to send {uppercase_name} from"
                ))
                .required(false),
        )
}

pub fn build_command() -> Command {
    let command = clap::command!();
    let mut transfer_subcommand = build_transfer_command();
    let mut stake_subcommand = build_stake_command();
    // TODO: this will break. how do i support multiple types here?
    #[allow(clippy::single_element_loop)]
    for value in &[eth::ETH] {
        let res = value.build_transfer_command(build_transfer_subcommand(value));
        match res {
            Ok(subcommand) => {
                transfer_subcommand = transfer_subcommand.subcommand(subcommand);
            }
            Err(_) => {
                // This coin does not support the given operation.
            }
        }

        let res = value.build_stake_command(build_stake_subcommand(value));
        match res {
            Ok(subcommand) => {
                stake_subcommand = stake_subcommand.subcommand(subcommand);
            }
            Err(_) => {
                // This coin does not support the given operation.
            }
        }
    }
    command
        .subcommand(transfer_subcommand)
        .subcommand(stake_subcommand)
}

pub fn handle_matches(_matches: clap::ArgMatches) -> Result<(), Box<dyn std::error::Error>> {
    Ok(())
}
